/* Copyright © 2013, Intel Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <limits.h>
#include <getopt.h>

#include <glaze.h>

typedef struct options {
	bool query;
	char *vendor;
	char *renderer;
	char *version;
	char *shading_language_version;
	char *extensions;
	char *extensions_whitelist;
	char *extensions_blacklist;
} options_t;

static void
export_options (options_t *options)
{
	if (options->query) {
		glaze_set_first_gl_call_callback ("glenv_query");
	}

	if (options->vendor)
		setenv ("GLENV_GL_VENDOR", options->vendor, 1);
	else
		unsetenv ("GLENV_GL_VENDOR");

	if (options->renderer)
		setenv ("GLENV_GL_RENDERER", options->renderer, 1);
	else
		unsetenv ("GLENV_GL_RENDERER");

	if (options->version)
		setenv ("GLENV_GL_VERSION", options->version, 1);
	else
		unsetenv ("GLENV_GL_VERSION");

	if (options->shading_language_version)
		setenv ("GLENV_GL_SHADING_LANGUAGE_VERSION",
			options->shading_language_version, 1);
	else
		unsetenv ("GLENV_GL_SHADING_LANGUAGE_VERSION");

	if (options->extensions)
		setenv ("GLENV_EXTENSIONS", options->extensions, 1);
	else
		unsetenv ("GLENV_EXTENSIONS");

	if (options->extensions_whitelist)
		setenv ("GLENV_EXTENSIONS_WHITELIST", options->extensions_whitelist, 1);
	else
		unsetenv ("GLENV_EXTENSIONS_WHITELIST");

	if (options->extensions_blacklist)
		setenv ("GLENV_EXTENSIONS_BLACKLIST", options->extensions_blacklist, 1);
	else
		unsetenv ("GLENV_EXTENSIONS_BLACKLIST");
}

static void
usage (void)
{
	printf ("Usage: glenv [OPTIONS...] <program> [program args...]\n"
		"\n"
		"Execute <program> with alternate OpenGL environment.\n"
		"\n"
		"Options:\n"
		"	-h, --help			Show this help message.\n"
		"	-q, --query			Query and report current environment\n"
		"					then terminate program.\n"
		"	--vendor=STR			Set GL_VENDOR to STR.\n"
		"	--renderer=STR			Set GL_RENDERER to STR.\n"
		"	--version=STR			Set GL_VERSION to STR.\n"
		"	--shading-language-version=STR	Set GL_SHADING_LANGUAGE_VERSION to STR.\n"
		"	--extensions=STR		Set GL_EXTENSIONS to STR.\n"
		"	--extensions-whitelist=STR	Remove from GL_EXTENSIONS all names\n"
		"					not appearing in STR (space-separated).\n"
		"	--extensions-blacklist=STR	Remove from GL_EXTENSIONS any names\n"
		"					appearing in STR (space-separated).\n");
}

enum {
	VENDOR_OPT = CHAR_MAX + 1,
	RENDERER_OPT,
	VERSION_OPT,
	SHADING_LANGUAGE_VERSION_OPT,
	EXTENSIONS_OPT,
	EXTENSIONS_WHITELIST_OPT,
	EXTENSIONS_BLACKLIST_OPT
};

static void
invalid_combination (const char *second, const char *first)
{
	fprintf (stderr, "Error: Option %s cannot be specified after %s\n",
		 second, first);
	exit (1);
}

int
main (int argc, char *argv[])
{
	int opt;
	options_t options = {
		.query = false,
		.vendor = NULL,
		.renderer = NULL,
		.version = NULL,
		.shading_language_version = NULL,
		.extensions = NULL,
		.extensions_whitelist = NULL,
		.extensions_blacklist = NULL
	};

	const char *short_options="hq";
	const struct option long_options[] = {
		{"help", no_argument, 0, 'h'},
		{"query", no_argument, 0, 'q'},
		{"vendor", required_argument, 0, VENDOR_OPT},
		{"renderer", required_argument, 0, RENDERER_OPT},
		{"version", required_argument, 0, VERSION_OPT},
		{"shading-language-version", required_argument, 0, SHADING_LANGUAGE_VERSION_OPT},
		{"extensions", required_argument, 0, EXTENSIONS_OPT},
		{"extensions-whitelist", required_argument, 0, EXTENSIONS_WHITELIST_OPT},
		{"extensions-blacklist", required_argument, 0, EXTENSIONS_BLACKLIST_OPT},
		{0, 0, 0, 0}
	};

	while (1)
	{
		opt = getopt_long (argc, argv, short_options, long_options, NULL);
		if (opt == -1)
			break;

		switch (opt) {
		case 'h':
			usage ();
			return 0;
		case '?':
			exit (1);
			break;
		case 'q':
			options.query = true;
			break;
		case VENDOR_OPT:
			options.vendor = optarg;
			break;
		case RENDERER_OPT:
			options.renderer = optarg;
			break;
		case VERSION_OPT:
			options.version = optarg;
			break;
		case SHADING_LANGUAGE_VERSION_OPT:
			options.shading_language_version = optarg;
			break;
		case EXTENSIONS_OPT:
			if (options.extensions_whitelist)
				invalid_combination ("--extensions", "--extensions-whitelist");
			if (options.extensions_blacklist)
				invalid_combination ("--extensions", "--extensions-blacklist");
			options.extensions = optarg;
			break;
		case EXTENSIONS_WHITELIST_OPT:
			if (options.extensions)
				invalid_combination ("--extensions-whitelist", "--extensions");
			if (options.extensions_blacklist)
				invalid_combination ("--extensions-whitelist", "--extensions-blacklist");
			options.extensions_whitelist = optarg;
			break;
		case EXTENSIONS_BLACKLIST_OPT:
			if (options.extensions)
				invalid_combination ("--extensions-blacklist", "--extensions");
			if (options.extensions_whitelist)
				invalid_combination ("--extensions-blacklist", "--extensions-whitelist");
			options.extensions_blacklist = optarg;
			break;
		default:
			fprintf (stderr, "Internal error: "
				 "unexpected getopt value: %d\n", opt);
			exit (1);
		}
	}

	if (optind >= argc) {
		fprintf (stderr, "Error: No program name provided, "
			 "see (glenv --help)\n");
		exit (1);
	}

	export_options (&options);

	glaze_execute (argc - optind, &argv[optind], "libglenv.so");

	/* If glaze_execute returns then something went wrong. */
	return 1;
}
